import os
import argparse

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import datasets, transforms
from torch.optim.lr_scheduler import StepLR


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        self.pool1 = nn.MaxPool2d(2, 1, 0)
        self.conv2 = nn.Conv2d(32, 32, 3, 1)
        self.pool2 = nn.MaxPool2d(2, 1, 0)
        self.conv3 = nn.Conv2d(32, 32, 3, 1)
        self.pool3 = nn.MaxPool2d(2, 1, 0)
        self.dropout1 = nn.Dropout2d(0.25)
        self.dropout2 = nn.Dropout2d(0.5)
        self.fc1 = nn.Linear(2592, 10)

    def forward(self, x):
        x = self.conv1(x)
        x = self.pool1(x)
        x = F.relu(x)
        x = self.conv2(x)
        x = self.pool2(x)
        x = F.relu(x)
        x = self.conv3(x)
        x = self.pool3(x)

        x = F.relu(x)
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = F.relu(x)
        output = F.log_softmax(x, dim=1)
        return output


def train(args, model, device, train_loader, optimizer, epoch):
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if batch_idx % args.log_interval == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch, batch_idx * len(data), len(train_loader.dataset),
                       100. * batch_idx / len(train_loader), loss.item()))


def test(args, model, device, test_loader):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += F.nll_loss(output, target, reduction='sum').item()  # sum up batch loss
            pred = output.argmax(dim=1, keepdim=True)  # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)

    print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n'.format(
        test_loss, correct, len(test_loader.dataset),
        100. * correct / len(test_loader.dataset)))


def main():
    # Training settings
    parser = argparse.ArgumentParser(description='PyTorch MNIST Example')
    parser.add_argument('--batch-size', type=int, default=64, metavar='N',
                        help='input batch size for training (default: 64)')
    parser.add_argument('--test-batch-size', type=int, default=1000, metavar='N',
                        help='input batch size for testing (default: 1000)')
    parser.add_argument('--epochs', type=int, default=1, metavar='N',
                        help='number of epochs to train (default: 14)')
    parser.add_argument('--lr', type=float, default=1.0, metavar='LR',
                        help='learning rate (default: 1.0)')
    parser.add_argument('--gamma', type=float, default=0.7, metavar='M',
                        help='Learning rate step gamma (default: 0.7)')
    parser.add_argument('--no-cuda', action='store_true', default=False,
                        help='disables CUDA training')
    parser.add_argument('--seed', type=int, default=1, metavar='S',
                        help='random seed (default: 1)')
    parser.add_argument('--log-interval', type=int, default=10, metavar='N',
                        help='how many batches to wait before logging training status')

    parser.add_argument('--save-model', action='store_true', default=True,
                        help='For Saving the current Model')
    args = parser.parse_args()
    use_cuda = not args.no_cuda and torch.cuda.is_available()

    torch.manual_seed(args.seed)

    device = torch.device("cuda" if use_cuda else "cpu")

    kwargs = {'num_workers': 1, 'pin_memory': True} if use_cuda else {}
    train_loader = torch.utils.data.DataLoader(
        datasets.MNIST('../data', train=True, download=True,
                       transform=transforms.Compose([
                           transforms.ToTensor(),
                           transforms.Normalize((0.1307,), (0.3081,))
                       ])),
        batch_size=args.batch_size, shuffle=True, **kwargs)
    test_loader = torch.utils.data.DataLoader(
        datasets.MNIST('../data', train=False, transform=transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ])),
        batch_size=args.test_batch_size, shuffle=True, **kwargs)

    model = Net().to(device)
    optimizer = optim.Adadelta(model.parameters(), lr=args.lr)

    scheduler = StepLR(optimizer, step_size=1, gamma=args.gamma)
    for epoch in range(1, args.epochs + 1):
        train(args, model, device, train_loader, optimizer, epoch)
        test(args, model, device, test_loader)
        scheduler.step()

    if args.save_model:
        ver = torch.__version__
        fp_ver = float(ver.split('.')[0]) + 0.1 * float(ver.split('.')[1])

        if fp_ver >= 1.6:
            if not os.path.exists('new_zipfile_serialization'):
                os.mkdir('new_zipfile_serialization/')
            if torch.cuda.is_available():
                if not os.path.exists('gpu'):
                    os.mkdir('new_zipfile_serialization/gpu')

                torch.save(model.state_dict(), "new_zipfile_serialization/gpu/mnist_state_dict.pt")
                torch.save(model, "new_zipfile_serialization/gpu/mnist_with_model_def.pt")
                torch.save({
                    'state_dict': model.state_dict(),
                    'optimizer': optimizer.state_dict(),
                }, "new_zipfile_serialization/gpu/mnist.pth.tar")

            else:
                if not os.path.exists('cpu'):
                    os.mkdir('new_zipfile_serialization/cpu')

                torch.save(model.state_dict(), "new_zipfile_serialization/cpu/mnist_state_dict.pt")
                torch.save(model, "new_zipfile_serialization/cpu/mnist_with_model_def.pt")
                torch.save({
                    'state_dict': model.state_dict(),
                    'optimizer': optimizer.state_dict(),
                }, "new_zipfile_serialization/cpu/mnist.pth.tar")

        else:
            if not os.path.exists('pickle_serialization'):
                os.mkdir('pickle_serialization/')
            if torch.cuda.is_available():
                if not os.path.exists('gpu'):
                    os.mkdir('pickle_serialization/gpu')

                torch.save(model.state_dict(), "pickle_serialization/gpu/mnist_state_dict.pt")
                torch.save(model, "pickle_serialization/gpu/mnist_with_model_def.pt")
                torch.save({
                    'state_dict': model.state_dict(),
                    'optimizer': optimizer.state_dict(),
                }, "pickle_serialization/gpu/mnist.pth.tar")

            else:
                if not os.path.exists('cpu'):
                    os.mkdir('pickle_serialization/cpu')

                torch.save(model.state_dict(), "pickle_serialization/cpu/mnist_state_dict.pt")
                torch.save(model, "pickle_serialization/cpu/mnist_with_model_def.pt")
                torch.save({
                    'state_dict': model.state_dict(),
                    'optimizer': optimizer.state_dict(),
                }, "pickle_serialization/cpu/mnist.pth.tar")


if __name__ == '__main__':
    os.environ['CUDA_VISIBLE_DEVICES'] = ' '
    print(torch.cuda.is_available())
    main()
